/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/


#ifndef MODALCV_NCC_H
#define MODALCV_NCC_H


#ifdef __cplusplus
extern "C" {
#endif


#include <stdint.h>


////////////////////////////////////////////////////////////////////////////////
// NCC on CVP
////////////////////////////////////////////////////////////////////////////////

#define MCV_NCC_MAX_TEMPLATES 750
// patch mode
#define MCV_NCC_PATCH_SIZE 64       // 8 bytes per pixel, 8x8 pixels, 64 bytes 
#define MCV_NCC_PATCH_SIDE_PIXELS 8 // 8 pixels

// general search window params
#define MCV_NCC_REF_SEARCH_SIZE_X 18
#define MCV_NCC_REF_SEARCH_SIZE_Y 18
// src search is only used from mcv_ncc_search_window in patch mode, otherwise we just the template at the index
#define MCV_NCC_SRC_SEARCH_SIZE_X 8
#define MCV_NCC_SRC_SEARCH_SIZE_Y 8


typedef enum mcv_ncc_mode{
    MCV_NCC_PATCH_MODE = 0, // NCC patch-based mode. Must assemble templates for NCC search.
    MCV_NCC_FRAME_MODE      // NCC frame-based mode. Can use previous frame as templates for NCC search.
} mcv_ncc_mode;

typedef enum mcv_ncc_score_type{
    MCV_NCC_BEST_SCORE = 0, // Gives output for the best NCC score only.
    MCV_NCC_ALL_SCORE       // Gives output for all NCC scores. Requires linearizing score output
} mcv_ncc_score_type;

typedef struct mcv_ncc_config_t{
    int width;
    int height;
    mcv_ncc_mode ncc_mode;
    mcv_ncc_score_type score_type;
    uint32_t max_templates;         // max is MCV_NCC_MAX_TEMPLATES
    int en_robustness;             // Specifies whether to enable robustness. By default, robustness is disabled (FALSE).
    int en_subpel;                 // Specifies whether to enable subpel. By default, subpel is enabled (TRUE).
} mcv_ncc_config_t;

typedef struct mcv_ncc_result_window{
    uint64_t match_loc_x;   // Top left of 18×18 window position x in the reference image.
    uint64_t match_loc_y;   // Top left of 18×18 window position y in the reference image.
    uint64_t src_loc_x;     // Top left of current source 8×8 window position x. For FRAME_MODE only, in PATCH_MODE, this parameter is copied from the search window src_loc_x
    uint64_t src_loc_y;     // Top left of current source 8×8 window position y. For FRAME_MODE only, in PATCH_MODE, this parameter is copied from the search window src_loc_y
    uint64_t robustness;    // Robustness value of NCC.
    uint64_t subpel;        // Indicates whether subpel was enabled (from configuration).
} mcv_ncc_result_window;

typedef struct mcv_ncc_search_window{
    uint64_t ref_loc_x;     // Top left of 18×18 window position x in the reference image.
    uint64_t ref_loc_y;     // Top left of 18×18 window position y in the reference image.
    uint64_t src_loc_x;     // Top left of current source 8×8 window position x.For FRAME_MODE only, otherwise ignored
    uint64_t src_loc_y;     // Top left of current source 8×8 window position y.For FRAME_MODE only, otherwise ignored
} mcv_ncc_search_window;

int mcv_ncc_init(mcv_ncc_config_t config);
int mcv_ncc_deinit(void);

// TODO  :  should there just be a wrapper here instead of seperate API calls?

// templates are 8x8 blocks, one uint8_t per pixel in raster scan order
// the maximum number of templates is MCV_NCC_MAX_TEMP, and we must have a search window for each
int mcv_ncc_patch_process(uint8_t* ref_img, uint8_t* templates, mcv_ncc_search_window* search_windows, int n_searches, mcv_ncc_result_window* results, int16_t* ncc_scores);
// same applies for search windows, but the src_loc_x/y MUST be filled and we just use curr_img instead of templates to find the matches
int mcv_ncc_frame_process(uint8_t* ref_img, uint8_t* curr_img, mcv_ncc_search_window* search_windows, int n_searches, mcv_ncc_result_window* results, int16_t* ncc_scores);

// this function is only needed if MCV_NCC_ALL_SCORE is selected
// otherwise, the scores are in linear order per template/request
// as of now, will still memcpy if using BEST_SCORE mode
int mcv_ncc_linearize_scores(int16_t* scores_in, int n_requests, int16_t* scores_out);

#ifdef __cplusplus
}
#endif

#endif // MODALCV_NCC_H